# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************

from .pipeline import Pipeline
from .logger import Logger
from .javascriptbuilder import JavascriptBuilderElement
from .jsonbundler import JSONBundlerElement
from .sequenceelement import SequenceElement
from .setheaderelement import SetHeaderElement
import importlib


class PipelineBuilder(object):
    """
    A PipelineBuilder generates a Pipeline object
    Before construction of the Pipeline, FlowElements are added to it
    There are also options for how JavaScript is output from the Pipeline

    """

    def __init__(self, settings= {}):
        """
        Pipeline Builder constructor.
        @type settings: dictionary
        @param settings : settings for the pipeline builder including:
        `add_javascript_builder (Bool)` - Whether to add the JavaScriptBuilder to the pipeline
        (default true)
        `javascript_builder_settings (dict)` - Settings for the JavaScriptBuilder engine 
        @rtype: PipelineBuilder
        `use_setheader_properties (Bool)` - Whether to add the SetHeaderElement to the pipeline
        (default true)
        @return: Returns a Pipeline Builder

        """

        self.flow_elements = []
        self.logger = Logger()

        if "add_javascript_builder" in settings:
            self.add_javaScriptbuilder = settings["add_javascript_builder"]
        else:
            self.add_javaScriptbuilder = True
          
        if "javascript_builder_settings" in settings:
            self.javascriptbuilder_settings = settings["javascript_builder_settings"]

        if "use_setheader_properties" in settings:
            self.use_setheader_properties = settings["use_setheader_properties"]
        else:
            self.use_setheader_properties = True


    def get_javascript_elements(self):
        """
        Adds the JavaScriptBuilder, JSONBundler and SequenceElement to the pipeline if
        If add_javascript_builder is set to true (the default)
        @rtype: list
        @return: Returns a list of FlowElements     
        """
        
        flow_elements = []

        if (self.add_javaScriptbuilder):

            flow_elements.append(SequenceElement())
            flow_elements.append(JSONBundlerElement())
    
            if (hasattr(self, "javascriptbuilder_settings")):
                flow_elements.append(JavascriptBuilderElement(self.javascriptbuilder_settings))
            else:
                flow_elements.append(JavascriptBuilderElement())
   
        return flow_elements

    def get_setheader_elements(self):
        """
        Adds the SetHeaderElement to the pipeline if
        If use_setheader_properties is set to true (the default)
        @rtype: list
        @return: Returns a list of FlowElements     
        """
        
        flow_elements = []

        if (self.use_setheader_properties):
            flow_elements.append(SetHeaderElement())
   
        return flow_elements

    def add(self, flow_element):
        """
        Add a flow_element to a list of flowElements be used in a pipeline.
        
        @type flow_element: FlowElement
        @param flow_element: flowElement to be added to the pipeline

        @rtype: PipelineBuilder
        @return: Returns the pipleine builder with the specified flowElement added to it's list of flowElements.

        """

        self.flow_elements.append(flow_element)

        return self

    def build(self):
        """
        Construct an immutable Pipeline using the list of flowElements, (Engines) and (Logger) currently in this Pipeline Builder.
        Call build after all items to be included in the pipeline have been added.

        @rtype: Pipeline
        @return: Returns a Pipeline

        """

        self.flow_elements.extend(self.get_javascript_elements())
        self.flow_elements.extend(self.get_setheader_elements())
        
        return Pipeline(self.flow_elements, logger=self.logger)

    def add_logger(self, logger):
        """
        Add an instance of the logger class to the pipeline.

        @type logger: Logger
        @param logger: Logger to be added to the pipeline

        @rtype: PipelineBuilder
        @return: Returns the pipeline builder with the specified Logger added.
        
        """

        self.logger = logger

        return self

    def build_from_configuration(self, config):
        """
        Build a pipeline from a configuration file

        @type config: dict
        @param config: pipeline configuration

        @rtype: Pipeline
        @return: Returns built pipeline
        
        """

        flow_elements = []

        for element in config["PipelineOptions"]["Elements"]:
            flow_element = importlib.import_module(element["elementPath"]).__getattribute__(element["elementName"])
            if("elementParameters" in element):
                flow_element = flow_element(**element["elementParameters"])
            else:
                flow_element = flow_element()
            flow_elements.append(flow_element)

        return Pipeline(flow_elements)




# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************


class Evidence:

    def __init__(self, flowdata):

        """!
        Constructor for Evidence container on a FlowData

        @type flowdata: FlowData
        @param flowdata: Parent FlowData

        """
        self.evidence = {}

        self.flowdata = flowdata

    def add(self, key, value):

        """!
        Add a single piece of evidence by its element and value

        @type key: string
        @param key: a flowElement's datakey

        @type value: mixed
        @param value: a piece of evidence

        """

        keep = False

        for flow_element in self.flowdata.pipeline.flow_elements:

            if flow_element.filter_evidence_key(key):
                keep = True

        if keep:
            self.evidence[key] = value

    def add_from_dict(self, evidence_dictionary):

        """!
        Helper function to set multiple pieces of evidence from a dict

        @type evidence_dictionary: dict
        @param evidence_dictionary: Dict of evidence

        """

        if not type(evidence_dictionary) is dict:
            self.flowdata.set_error("core", "Must pass valid dictionary.")

        for key, value in evidence_dictionary.items():
            self.add(key, value)

    def get(self, key):

        """!
        Get a piece of evidence by key

        @type key: string
        @param key: A FlowElement's datakey

        @rtype: dict
        @return: A piece of evidence

        """

        if key in self.evidence:

            return self.evidence[key]

        else:

            return

    def get_all(self):

        """!
        Get all evidence

        @rtype: dict
        @return: Returns everything in this Evidence.evidence
        """

        return self.evidence




# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************

from .evidence_keyfilter import EvidenceKeyFilter


class FlowElement(object):

    """!
    A FlowElement is placed inside a Pipeline
    It receives Evidence via a FlowData object
    It uses this to optionally create ElementData on the FlowData
    Any errors in processing are caught in the FlowData's errors object
    
    """

    def __init__(self):

        """!
        List of Pipelines the FlowElement has been added to 

        """
        self.pipelines = []
        self.properties = {}
        self.datakey = ""

        # Allow elements to be excluded from messages listing
        # all flowelements. For system elements like the JavaScriptBuilder
        
        self.exclude_from_messages = False

    def process(self, flowdata):

        """!
        Function for getting the FlowElement's EvidenceKeyFilter
        Used by the filterEvidence method

        @type flowdata: FlowData
        @param flowdata: FlowData to be processed

        @rtype: mixed
        @return: Returns whatever the self.processInternal method is set to return

        """

        return self.process_internal(flowdata)

    def on_registration(self, pipeline):

        """!
        Function called when an element is added to the pipeline. 
        Used for example, for elements that depend on other elements in a pipeline

        @type pipeline: Pipeline
        @param pipeline: Pipeline the element has been added to

        """

        pass

    def get_evidence_key_filter(self):

        """!
        Filter FlowData evidence using the FlowElement's EvidenceKeyFilter

        @rtype: EvidenceKeyFilter
        @return: Returns an EvidenceKeyFilter

        """

        return EvidenceKeyFilter()

    def filter_evidence(self, flowdata):

        """!
        Filter FlowData evidence using the FlowElement's EvidenceKeyFilter

        @type flowdata: FlowData
        @param flowdata: a FlowData that has some Evidence set

        @rtype: dict
        @return: Returns a dictionary of evidence that has passed the filter

        """

        filter = self.get_evidence_key_filter()

        return filter.filter_evidence(flowdata.evidence.get_all())

    def filter_evidence_key(self, key):

        """!
        Filter FlowData.evidence using the flowElement's EvidenceKeyFilter
        with the property key of evidence of interest.

        @type key: string
        @param key: the property key being sought within FlowData.evidence

        @rtype: dict
        @return: Returns a dictionary containing the property key and the evidence related to it as its value

        """

        filter = self.get_evidence_key_filter()

        return filter.filter_evidence_key(key)

    def process_internal(self, flowdata):

        """!
        The method behind FlowElement.Process - it is called by the process() function.
        It is usually overridden by specific flowElements to do their core work.

        @type flowdata: FlowData
        @param flowdata: FlowData to be processed

        @rtype: bool
        @return: Returns True

        """

        return True

    def get_properties(self):

        """!
        Get the FlowElement.properties of a FlowElement.
        
        This is usually overridden by specific flowElements.

        @rtype: DataPropertyDictionary
        @return: Returns dictionary of the FlowElement's properties

        """

        return {k.lower(): v for k, v in self.properties.items()}



 # *********************************************************************
 # This Original Work is copyright of 51 Degrees Mobile Experts Limited.
 # Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
 # Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
 #
 # This Original Work is licensed under the European Union Public Licence (EUPL) 
 # v.1.2 and is subject to its terms as set out below.
 #
 # If a copy of the EUPL was not distributed with this file, You can obtain
 # one at https://opensource.org/licenses/EUPL-1.2.
 #
 # The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
 # amended by the European Commission) shall be deemed incompatible for
 # the purposes of the Work and the provisions of the compatibility
 # clause in Article 5 of the EUPL shall not apply.
 # 
 # If using the Work as, or as part of, a network application, by 
 # including the attribution notice(s) required under Article 5 of the EUPL
 # in the end user terms of the application under an appropriate heading, 
 # such notice(s) shall fulfill the requirements of that article.
 # ********************************************************************


class AspectPropertyValue(object):
    
    """!
    An AspectPropertyValue is a wrapper for a value
    It lets you check this wrapper has a value inside it
    If not value is set, a specific no value message is returned
    """

    def __init__(self, no_value_message = None, value = "noValue"):
    
        """!
        Constructor for AspectPropertyValue

        @type value: mixed
        @param value: the property value to store

        @type noValueMessage: string
        @param noValueMessage: if there is no value, the reason for there not being one
        
        """

        self.noValueMessage = no_value_message

        if(value != "noValue"):
            self.__value = value
            self.__hasValue = True
        else:
            self.__hasValue = False

    def has_value(self):

        """!
        Check if the AspectPropertyValue wrapper has a value inside it

        @rtype bool
        @return whether there is a value
        
        """

        return self.__hasValue

    def value(self):

        """!
        Get the value out of the AspectPropertyValue wrapper
        Note that this will return an error (including no_value_message if set)
        if there is no value. So check has_value() first to be sure.

        @rtype mixed
        @return the stored value
        
        """
        
        if self.__hasValue:
            return self.__value
        else:
            raise Exception(self.no_value_message())

    def no_value_message(self):

        """!
        If there is no value, get the reason for there not being a value

        @rtype string
        @return the no value message
        
        """

        return self.noValueMessage



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************


from .flowelement import FlowElement
from .evidence_keyfilter import EvidenceKeyFilter
from .elementdata_dictionary import ElementDataDictionary
import os
import json

try:
    #python2
    from urllib import urlencode
except ImportError:
    #python3
    from urllib.parse import urlencode

import chevron
from jsmin import jsmin


class JavaScriptBuilderEvidenceKeyFilter(EvidenceKeyFilter):

    def filter(self, key):
        if "query" in key:
            return True

        if key == "header.host" or key == "header.protocol":
            return True
            
        return False


class JavascriptBuilderElement(FlowElement):

    """!
    The JavaScriptBuilder aggregates JavaScript properties
    from FlowElements in the Pipeline. This JavaScript also (when needed)
    generates a fetch request to retrieve additional properties
    populated with data from the client side
    It depends on the JSON Bundler element (both are automatically
    added to a Pipeline unless specifically removed) for its list of properties.
    The results of the JSON Bundler should also be used in a user-specified
    endpoint which retrieves the JSON from the client side.
    The JavaScriptBuilder is constructed with a url for this endpoint.

    """

    def __init__(self, settings = {} ):

        """!
        JavaScriptBuilder constructor.

        * @param {dict} options options object
        * @param {string} options.obj_name the name of the client
        * side object with the JavaScript properties in it ('fod' by default)
        * @param {string} options.protocol The protocol ("http" or "https")
        * used by the client side callback url.
        * This can be overriden with header.protocol evidence
        * @param {string} options.host The host of the client side
        * callback url. This can be overriden with header.host evidence.
        * @param {string} options.endpoint The endpoint of the client side
        * callback url
        * @param {boolean} options.enable_cookies whether cookies should be enabled
        * @param {boolean} options.minify Whether to minify the JavaScript

        """

        super(JavascriptBuilderElement, self).__init__()
        
        self.settings = {}

        self.settings['_objName'] = settings["obj_name"] if "obj_name" in settings else 'fod'
        self.settings['_protocol'] = settings["protocol"] if "protocol" in settings else None
        self.settings['_host'] = settings["host"] if "host" in settings else None
        self.settings['_endpoint'] = settings["endpoint"] if "endpoint" in settings else ''
        self.settings['_enableCookies'] = settings["enable_cookies"] if "enable_cookies" in settings else True

        self.minify = settings["minify"] if "minify" in settings else True

        self.datakey = "javascriptbuilder"

        self.exclude_from_messages = True

        # Load template file contents into memory

        template = os.path.dirname(os.path.abspath(__file__)) + "/JavaScriptResource.mustache"

        f = open(template, "r")
        self.template = f.read()
        f.close()

    
    def get_evidence_key_filter(self):

        """!
        
        The JavaScriptBuilder captures query string evidence and
        headers for detecting whether the request is http or https
    
        """
   
        return JavaScriptBuilderEvidenceKeyFilter()


    def process_internal(self, flowdata):

        """!
        The JavaScriptBundler collects client side javascript to serve.

        @type flowdata: FlowData
        @param flowdata: The FlowData

        """
    
        variables = {}

        for key, value in self.settings.items():
            variables[key] = value

        variables["_jsonObject"] = json.dumps(flowdata.jsonbundler.json)

        # Generate URL and autoUpdate params

        host = self.settings["_host"]
        protocol = self.settings["_protocol"]

        if not protocol:
            # Check if protocol is provided in evidence
            if flowdata.evidence.get("header.protocol"):
                protocol = flowdata.evidence.get("header.protocol")
            
        if not protocol:
            protocol = "https"
      

        if not host:
            # Check if host is provided in evidence

            if flowdata.evidence.get("header.host"):
                host = flowdata.evidence.get("header.host")

        variables["_host"] = host
        variables["_protocol"] = protocol

        if variables["_host"] and variables["_protocol"] and variables["_endpoint"]:

            variables["_url"] = variables["_protocol"] + "://" + variables["_host"] + variables["_endpoint"]

            # Add query parameters to the URL

            query_params = self.get_evidence_key_filter().filter_evidence(flowdata.evidence.get_all())

            query = {}
 
            for param, paramvalue in query_params.items():

                paramkey = param.split(".")[1] 

                query[paramkey] = paramvalue
  
            url_query = urlencode(query)
            
            # Does the URL already have a query string in it?
    
            if "?" not in variables["_url"]: 
                variables["_url"] += "?"
            else:
                variables["_url"] += "&"
            
            variables["_url"] += url_query

            variables["_updateEnabled"] = True
        else:
            variables["_updateEnabled"] = False
        

        # Use results from device detection if available to determine
        # if the browser supports promises.
        
        try:
            flowdata.get("device").get("promise")
            variables["_supportsPromises"] = flowdata.device.promise.value == True
        except Exception:
            variables["_supportsPromises"] = False
       
        # Check if any delayedproperties exist in the json

        variables["_hasDelayedProperties"] = True if "delayexecution" in variables["_jsonObject"] else False
         
        output = chevron.render(self.template, variables)
        
        if self.minify:
            # Minify the output
            output = jsmin(output)
        
        data = ElementDataDictionary(self, {"javascript": output})

        flowdata.set_element_data(data)

        return



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************

from .flowelement import FlowElement
from .elementdata_dictionary import ElementDataDictionary

class JSONBundlerElement(FlowElement):

    """!
    The JSONBundler aggregates all properties from FlowElements into a JSON object
    It is used for retrieving via an endpoint from the client
    side via the JavaScriptBuilder and also used inside the
    JavaScriptBuilder itself to pass properties to the client side.
    Both this and the JavaScriptBuilder element are automatically
    added to a Pipeline unless specifically ommited in the PipelineBuilder

    """

    def __init__(self):

        super(JSONBundlerElement, self).__init__()

        self.datakey = "jsonbundler"

        self.exclude_from_messages = True

        self.properties = {"json" : { "type": "dict"} }

        self.property_cache = {}

    def process_internal(self, flowdata):

        """!
        
        The JSONBundler extracts all properties from a FlowData and serializes them into JSON
        @type flowdata: FlowData
        @param flowdata: A FlowData
    
        """
   
        # Get every property on every FlowElement
        # Storing JavaScript properties in an extra section

        output = {"javascriptProperties": [] }

        if len(self.property_cache):
            property_cache_set = True
        else:
            property_cache_set = False
        
        for flow_element in flowdata.pipeline.flow_elements:

            if (flow_element.datakey == "jsonbundler" or
                flow_element.datakey == "sequence" or 
                flow_element.datakey == "javascriptbuilder" or
                flow_element.datakey == "set-headers"):
                continue
            
            properties = flow_element.get_properties()

            if not property_cache_set:

                delay_execution_list = []
                delayed_evidence_properties = {}

                # Loop over all properties and see if any have delay execution set to true

                for propertykey, propertymeta in properties.items():
                    
                    if "delayexecution" in propertymeta and propertymeta["delayexecution"] == True:
                        delay_execution_list.append(propertykey)

                """
                Loop over all properties again and see if any have evidenceproperties which
                have delayedExecution set to true
                """
                
                for propertykey, propertymeta in properties.items():

                    if("evidenceproperties" in propertymeta):

                        delayed_evidence_properties_list = list(filter(lambda x: x in delay_execution_list, propertymeta["evidenceproperties"]))

                        if len(delayed_evidence_properties_list):
                            
                            delayed_evidence_properties[propertykey] = list(map(lambda x:
                                flow_element.datakey + '.' + x, delayed_evidence_properties_list))


                self.property_cache[flow_element.datakey] = {
                    "delayExecutionList": delay_execution_list,
                    "evidenceProperties": delayed_evidence_properties
                }

            property_cache = self.property_cache[flow_element.datakey]

            # Create empty area for FlowElement properties to go

            output[flow_element.datakey] = {}

            for propertykey, propertymeta in properties.items():
                value = None
                null_reason = "Unknown"

                # Check if property has delayed execution and set in JSON if yes


                if propertykey in property_cache["delayExecutionList"]:
                    output[flow_element.datakey][propertykey.lower() + "delayexecution"] = True
                

                # Check if property has any delayed execution evidence properties and set in JSON if yes

                if propertykey in property_cache["evidenceProperties"]:
                    output[flow_element.datakey][propertykey.lower() + 'evidenceproperties'] = property_cache["evidenceProperties"][propertykey]
                

                try:

                    value_container = flowdata.get(flow_element.datakey).get(propertykey)

                    # Check if value is of the aspect property value type

                    if isinstance(value_container, object) and hasattr(value_container, "has_value" ):
                    
                        # Check if it has a value

                        if value_container.has_value():
                            value = value_container.value()
                        else:
                            value = None
                            null_reason = value_container.no_value_message()
                        
                    # Check if list of aspect property values

                    elif isinstance(value_container, list) and isinstance(value_container[0], object):

                        output = []

                        for item in value_container:
                            if item.has_value():
                                output.append(item.value())
                            else:
                                null_reason = item.no_value_message

                        value = output

                    else:

                        # Standard value
                        value = value_container
                    
                except:
                    # Catching missing property exceptions and other errors

                    continue
        
                output[flow_element.datakey.lower()][propertykey.lower()] = value
                if value is None:
                    output[flow_element.datakey.lower()][propertykey.lower() + "nullreason"] = null_reason

                sequence = flowdata.evidence.get("query.sequence")

                if sequence is None or sequence < 10: 

                    if "type" in propertymeta and propertymeta["type"].lower() == "javascript":

                        output["javascriptProperties"].append(flow_element.datakey.lower() + "." + propertykey.lower())
                 
        data = ElementDataDictionary(self, {"json": output})

        flowdata.set_element_data(data)

        return



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************

from .flowdata import FlowData
from .logger import Logger

class Pipeline:

    """!
    Pipeline holding a list of FlowElements for processing,
    can create FlowData that will be passed through these,
    collecting ElementData
    Should be constructed through the PipelineBuilder class

    """

    def __init__(self, flow_elements, logger=Logger()):

        """!
        Pipeline constructor.

        @type flow_elements: list[FlowElement]
        @param flow_elements: A list of flowElements

        @type logger: Logger
        @param logger: A logger to attach to the pipeline

        @rtype: Pipeline
        @return: Returns a Pipeline

        """

        self.flow_elements = flow_elements

        self.logger = logger

        self.flow_elements_list = {}

        self.flow_elements_display_list = []

        for flow_element in flow_elements:

            # Notify element that it has been registered in the pipeline
            flow_element.on_registration(self)

            self.flow_elements_list[flow_element.datakey] = flow_element

            flow_element.pipelines.append(self)

            if not flow_element.exclude_from_messages:
                self.flow_elements_display_list.append(flow_element.datakey)

    def create_flowdata(self):

        """!
        Create a FlowData based on what's in the pipeline
        
        @rtype: FlowData
        @return: Return a FlowData

        """

        return FlowData(self)

    def log(self, level, message):

        """!
        Log a message using the Logger.log of the pipeline's Logger.

        @type level: string
        @param level: level of log message

        @type message: string
        @param message: Returns content of log message

        """

        self.logger.log(level, message)

    def get_element(self, key):

        """!
        Get a flowElement by its name.

        @type key: string
        @param key: name of flowElement

        @rtype: FlowElement
        @return: Returns the FlowElement indicated

        """

        return self.flow_elements_list[key]

    def element_not_found(self, element, flowdata):
        """!
        Trigger error when an element cannot be found in the pipeline

        @type element: string
        @param element: name of flowelement 
        @type flowdata: flowdata
        @param element: flowdata being processsed 
        @rtype: Exception
        @return: Returns exception

        """

        raise Exception("There is no element data for " + element + " against this flow data. Available element data keys are: " + str(flowdata.pipeline.flow_elements_display_list))


    def get_properties(self):

        """!
        Get all properties of all flowElements in the pipeline.

        @rtype: dict of {string : DataPropertyDictionary}
        @return: Returns dictionary of all properties in a pipeline keyed by each flowElement's FlowElement.datakey.

        """

        output = {}

        for flow_element in self.flow_elements:
            properties = flow_element.get_properties()

            output[flow_element.datakey] = properties

        return output



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************

from datetime import datetime
import logging

class Logger:
    
    """!
    Logging for a Pipeline

    """

    def __init__(self, min_level="error", settings = {}):

        """!
        Create a logger

        @type minLevel: str|list
        @param minLevel: Logging level ("debug", "info", "warning", "error", "critical")

        """

        self.allowed_levels = ["debug", "info", "warning", "error", "critical"]

        self.min_level = min_level

        # enable this level in logging to be output (system default is 'warning')

        logging.basicConfig(level=getattr(logging, min_level.upper()))

        self.min_level = self.allowed_levels.index(str(min_level).lower())
        
        self.settings = settings

    def log(self, level, message):

        """!
        Log a message

        @type level: string
        @param level: The level of log message

        @type message: string
        @param message: The content of log message

        """

        level_index = self.allowed_levels.index(str(level).lower())

        if level_index >= self.min_level:
            now = datetime.now()

            log = {"time": now.strftime("%Y-%m-%d, %H:%M:%S"), "level": level, "message": message}
            self.log_internal(level, log)


    def log_internal(self, level, log):

        """!
        Internal logging function overridden by specific loggers

        @type level: string
        @param level: The level of log message

        @type log: dict
        @param log: The body of log entry
        
        """

        if level == 'debug':
            logging.debug(log)

        elif level == 'info':
            logging.info(log)

        elif level == 'warning':
            logging.warning(log)

        elif level == 'error':
            logging.error(log)

        elif level == 'critical':
            logging.critical(log)

        return



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************

from .evidence_keyfilter import EvidenceKeyFilter


class BasicListEvidenceKeyFilter(EvidenceKeyFilter):

    """!
    An instance of EvidenceKeyFilter that uses a simple array of keys
    Evidence not using these keys is filtered out

    """

    def __init__(self, keys_list):
        
        """!
        BasicListEvidenceKeyFilter constructor

        @type keysList: dict 
        @param keysList: a list of keys to keep.

        """

        # lowercase keys list
        self.list = [item.lower() for item in keys_list]

    def filter_evidence_key(self, key):

        """!
        Check if an evidence key is present in a filter list.

        @type key: string
        @param key: to check in the filter

        @rtype: bool
        @return: Is this key in the filter's keys list?

        """
        
        if key.lower() in self.list:
            return True
        else:
            return False



from flask import request

def webevidence(request):

    """!
    Get evidence from a web request (gets headers, cookies and query parameters)
    
    @type request: Request 
    @param request: A Request object
    @rtype dict
    @return A dictionary of web evidence that can be using in flowdata.evidence.add_from_dict()

    """

    webevidence = {}

    for header in request.headers:
        webevidence["header." + header[0].lower()] = header[1]

    for cookieKey, cookieValue in request.cookies.items():
        webevidence["cookie." + cookieKey] = cookieValue

    for query,value in request.args.items():

        webevidence["query." + query] = value
    
    webevidence["server.client-ip"] =  request.remote_addr

    webevidence["server.host-ip"] =  request.host

    if (request.is_secure):
        webevidence["header.protocol"] = "https"
    else:
        webevidence["header.protocol"] = "http"

    return webevidence

def set_response_header(flowData, response):
    
    """!
    Set UACH response header in web response (sets Accept-CH header in response)
    
    @type response: Response 
    @param response: A Response object
    @param response_header_dict: Dictionary containing response header key and values to be set
    @rtype response
    @return A response object containing headers with non null values in response

    """
    
    # Get response headers dictionary containing key values to be set in response  
    response_header_dict = flowData["set-headers"]["responseheaderdictionary"]
    for response_key, response_value in response_header_dict.items():
        response_value = response_value.replace(",", ", ")
        if response_value != "":
            response.headers[response_key] = response_value

    return response


# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************* 


from .flowelement import FlowElement
from .evidence_keyfilter import EvidenceKeyFilter
import uuid

class SequenceElementEvidenceKeyFilter(EvidenceKeyFilter):

    def filter(self, key):
    
        if (key == "query.sequence" or key == "query.session-id"):
            return True
            
        return False


class SequenceElement(FlowElement):

    """!
    The SequenceElement stores session data regarding
    requests for client side JavaScript from the JavaScript
    created by a Pipeline's JavaScriptBuilder
    If a Pipeline is constructed with the JavaScript elements
    enabled this is added automatically along with the JavaScriptBuilder
    and JSONBundler.
    """

    def __init__(self):

        super(SequenceElement, self).__init__()

        self.datakey = "sequence"
        self.exclude_from_messages = True


    def get_evidence_key_filter(self):
    
        """!
        The SequenceElement uses query.sequence and query.session-id evidence
        """
  
        return SequenceElementEvidenceKeyFilter()

    def process_internal(self, flowdata):

        """!
        The SequenceElement stores session data for requests for JavaScript
    
        @type flowdata: FlowData
        @param flowdata: The FlowData

        """
    
        if flowdata.evidence.get("query.session-id"):
            
            # Get current sequence number
    
            sequence = flowdata.evidence.get("query.sequence")
      
            if sequence:
                sequence = int(sequence)
            else:
                sequence = 1
           
      
            flowdata.evidence.add("query.sequence", sequence + 1)

        else:
            flowdata.evidence.add(
                "query.session-id",
                str(uuid.uuid4())
            )
      
            flowdata.evidence.add("query.sequence", 1)



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************


class EvidenceKeyFilter:

    """!
    An EvidenceKeyFilter is attached to a FlowElement and is used to check
    if a specific evidence key is needed by it. This allows for keys that are
    not used by any FLowElement in a pipeline to be discarded, and on a FlowElement
    level allows for caching of data / memoization based on the evidence a FlowElement
    requires.

    """

    def filter_evidence(self, evidence_key_object):

        """!
        Filter evidence from a dictionary of evidence keys/values.
        Runs filterEvidenceKey on each key in the dictionary

        @type evidence_key_object: dict
        @param evidence_key_object: Evidence dictionary contents

        @rtype: dict
        @return: EReturns filtered evidence dictionary contents
        """

        filtered = {}

        for key, value in evidence_key_object.items():

            if self.filter_evidence_key(key):

                filtered[key] = value

        return filtered

    def filter_evidence_key(self, key):

        """!
        See if a property key should be in the filtered evidence

        @type key: string
        @param key: property name

        @rtype: bool
        @return: Returns is this be filtered out or not?

        """

        return True



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************

from .elementdata import ElementData


class ElementDataDictionary(ElementData):

    """!
    An extension of ElementData with dictionary object storage / lookup

    """

    def __init__(self, flow_element, contents):
        
        """!
        Constructor for Element Data Dictionary

        @type flowElement: FlowElement
        @param flowElement: FlowElement that creates the data to be stored

        @type contents: dict
        @param contents: Dictionary contents

        """

        super(ElementDataDictionary, self).__init__(flow_element)
        self.contents = {}

        for key, value in contents.items():
            self.contents[key.lower()] = value

        self.flow_element = flow_element

        ElementData(flow_element)

    def as_dictionary(self):
        
        """!
        Get the values contained in the ElementData instance as a dictionary
        of keys and values.

        @rtype: dict
        @return: Returns a dictionary of items in an ElementData

        """

        return self.contents

    def get_internal(self, key):

        """!
        Internal getter for ElementDataDictionary.contents

        @type key: string
        @param key: Key of an item in the ElementDataDictionary.

        @rtype: mixed
        @return: The data keyed under that property

        """

        if key in self.contents:

            return self.contents[key]

        else:

            return



 ##********************************************************************
 # This Original Work is copyright of 51 Degrees Mobile Experts Limited.
 # Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
 # Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
 #
 # This Original Work is licensed under the European Union Public Licence (EUPL)
 # v.1.2 and is subject to its terms as set out below.
 #
 # If a copy of the EUPL was not distributed with this file, You can obtain
 # one at https://opensource.org/licenses/EUPL-1.2.
 #
 # The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
 # amended by the European Commission) shall be deemed incompatible for
 # the purposes of the Work and the provisions of the compatibility
 # clause in Article 5 of the EUPL shall not apply.
 #
 # If using the Work as, or as part of, a network application, by
 # including the attribution notice(s) required under Article 5 of the EUPL
 # in the end user terms of the application under an appropriate heading,
 # such notice(s) shall fulfill the requirements of that article.
 # ******************************************************************** 

 # Messages which may be reused for various property exceptions.

class Messages():

    # Property does not start with SetHeader. This takes the name of a property
    # as a format argument.

    PROPERTY_NOT_SET_HEADER = "Property Name '{}' does not start with 'SetHeader'."

    # Property Name is not in the valid format. This takes the property name 
    # as format argument.

    WRONG_PROPERTY_FORMAT = \
        "Property Name '{}' is not in the expected format i.e. SetHeader[Component][HeaderName]. "
    
    # Element not found in flowData. This takes the element datakey 
    # as format argument.
    ELEMENT_NOT_FOUND = \
        "Element '{}' is not present in the FlowData. "

    # Property not found in flowData. This takes the element datakey
    # and property names as format arguments.
    PROPERTY_NOT_FOUND = \
        "Property '{}' is not present in the FlowData against '{}' ElementData. "


import setuptools
import os
import io

# Read a text file and return the content as a string.
def read(file_name):

    """Read a text file and return the content as a string."""
    try:
        with io.open(
            os.path.join(os.path.dirname(__file__), file_name), encoding="utf-8"
        ) as f:
            return f.read()
    except:
        return ""
    
setuptools.setup(
    name="fiftyone_pipeline_core",
    version=read("version.txt"),
    author="51Degrees",
    author_email="support@51degrees.com",
    url="https://51degrees.com/",
    description=("The 51Degrees Pipeline API is a generic web request intelligence and data processing solution with the ability to add a range of 51Degrees and/or custom plug ins (Engines). "
    "This package definds the essential components of the Pipeline API such as flow elements, flow data and evidence. It also packages together JavaScript served by a pipeline and allows for client side requests for additional data populated by evidence from the client side."),
    long_description=read("readme.md"),
    long_description_content_type='text/markdown',
    python_requires='>=3.5',
    packages=["fiftyone_pipeline_core"],
    install_requires=[
          'chevron',
          'jsmin'
    ],
    license="EUPL-1.2",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python",
        "License :: OSI Approved :: European Union Public Licence 1.2 (EUPL 1.2)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    include_package_data=True
)



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************


class ElementData(object):

    """!
    Core ElementData class
    Stores information created by a FlowElement based on FlowData.
    Stored in FlowData

    """

    def __init__(self, flow_element):

        """!
        Constructor for ElementData

        @type flow_element: FlowElement
        @param flow_element: FlowElement that data has been created by

        """

        self.flow_element = flow_element

    def get(self, key):

        """!
        Get a value from the elementData contents
        Proxy for the ElementData.getInternal method

        @type key: str
        @param: key

        @rtype: mixed
        @return: Returns specified property from the data

        """

        return self.get_internal(key.lower())

    def __getattr__(self, key):

        """!
        Magic getter for a property from the data
        Allowing  user to write for example Data.IsMobile rather than Data.get("IsMobile)

        @type key: str
        @param key: property

        @rtype: mixed
        @return: Returns specified property from the data
        """

        return self.get(key)

    def __getitem__(self, flow_element_key):
    
        """!
        Magic method in Python, which when used in a class, allows its instances to use the [] (indexer) operators.
        For example, instead of `flowdata.get("device")` you can use `flowdata["device"]`

        @type flow_element_key: string
        @param flow_element_key: datakey of the FlowElement that created the data of interest

        @rtype: ElementData
        
        """

        return self.get(flow_element_key)
		
    def get_internal(self, key):
        
        """!
        Returns the requested property from the data
        Overridden by specific ElementData instances.

        @type key: str
        @param key: property

        @rtype: mixed
        @return: Returns Requested property from data
        """

        return

    def get_properties(self):

        """!
        Proxy to the data's flowElement properties.
        Gets the FlowElement.properties of the parent FlowElement.

        @rtype: DataPropertyDictionary
        @return: Returns dictionary of the parent FlowElement's properties
        """

        return self.flow_element.get_properties()



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************

from .flowelement import FlowElement
from .elementdata_dictionary import ElementDataDictionary
from .messages import Messages
import re

class SetHeaderElement(FlowElement):

    """!
    Set response headers element class. This is used to get response
    headers based on what the browser supports. For example, newer
    Chrome browsers support the Accept-CH header.

    """

    def __init__(self):

        super(SetHeaderElement, self).__init__()

        self.datakey = "set-headers"

        self.properties = {"responseheaderdictionary" : { "type": "dict"} }

        self.setheader_properties = {}
		
        self.exclude_from_messages = True

    def process_internal(self, flowdata):
    
        """!
        
        Add the response header dictionary to the FlowData.
        @type flowdata: FlowData
        @param flowdata: A FlowData
    
        """

        if not self.setheader_properties:
            self.setheader_properties = self.get_setheader_properties_pipeline(flowdata.pipeline)
        
        response_headers = self.get_response_header_value(flowdata, self.setheader_properties)

        data = ElementDataDictionary(self, {"responseheaderdictionary": response_headers})

        flowdata.set_element_data(data)

        return

    def get_setheader_properties_pipeline(self, pipeline):
        
        """!
        Get All the properties starting with SetHeader string from pipeline
        
        @param pipeline: A Pipeline object
        @rtype dict
        @return A dictionary object containing SetHeader properties list against flowElement

        """
        
        setHeader_properties_dict = {}

        # Loop over each flowElement in pipeline to check SetHeader properties
        for flow_element in pipeline.flow_elements:

            # Get the properties against the flowElement
            properties = flow_element.get_properties()

            setHeader_element_list = []

            # Loop over each flowElement property
            for propertykey, propertymeta in properties.items():

                # Check if the property starts with SetHeader
                if("setheader" in propertykey):
                    setHeader_element_list.append(propertymeta["name"])

            # Add SetHeader element list in dict against flowElement.datakey as key
            if setHeader_element_list:
                setHeader_properties_dict[flow_element.datakey] = setHeader_element_list
        
        return setHeader_properties_dict

    def get_response_header_value(self, flowData, setHeader_properties_dict):

        """!
        Get response header value using set header properties from FlowData
        
        @type flowdata: FlowData 
        @param flowdata: A processed FlowData object containing setheader properties
        @param setHeader_properties_dict: A processed FlowData object containing setheader properties
        @rtype dict
        @return A dictionary object containing SetHeader properties list against flowElement

        """

        response_headers_dict = {}

        # Loop over all the flowElements to process Set Header properties for User Agent Client Hints
        for element_datakey, setHeader_element_list in setHeader_properties_dict.items():

            # Loop over each setHeader property of the element  
            for setHeader_property in setHeader_element_list:

                # Get response header key to be set in response
                response_header = self.get_response_header_name(setHeader_property)

                # Get SetHeader property value from elementData
                setHeader_value = self.get_property_value(flowData, element_datakey, setHeader_property)

                # Process and Add property in the dict against response header key
                if(response_header in response_headers_dict):
                    response_header_value = response_headers_dict[response_header]
                    if(response_header_value == ""):
                        response_header_value = setHeader_value
                    else:
                        if(setHeader_value != ""):
                            response_header_value = response_header_value + "," + setHeader_value
                    response_headers_dict[response_header] = response_header_value
                else:
                    response_headers_dict[response_header] = setHeader_value

        return response_headers_dict

    def get_property_value(self, flowData, element_key, property_key):

        """!
        Try to get the value for the given element and property.
        If the value cannot be found or is null/unknown, then ""
        will be returned.

        @param flowData: A processed FlowData instance to get the value from.
        @param element_key: Key for the element data to get the value from
        @param property_key: Name of the property to get the value for.
        @rtype String
        @return value or empty string

        """

        value = ""
        
        try:
            # Get the elementData from flowData that contains required property.
            element_data = flowData.get(element_key)
        except Exception:
            print(Messages.ELEMENT_NOT_FOUND.format(element_key))
            return ""

        try:
            property = element_data[property_key.lower()]
        except:
            print(Messages.PROPERTY_NOT_FOUND.format(property_key, element_key))
            return ""

        if(property.has_value() and property.value() not in ["Unknown", "noValue"]):
            value = property.value()
        else:
            value = ""

        return value

    def get_response_header_name(self, property_key):
        
        """!
        Determines which response header the property value will be appended to by 
        stripping the 'SetHeader' string and the 'Component Name' from the property name.
        
        @param property_key: Key for SetHeaderAcceptCH property
        @rtype String
        @return Response Header name

        """

        actual_property_name = property_key
        response_header = ""

		# Check if property name starts with SetHeader.
		# If Yes, Discard SetHeader from property name.
        if ("SetHeader" not in property_key[0:9]):
            raise Exception(Messages.PROPERTY_NOT_SET_HEADER.format(actual_property_name))
        else:
            # Discard SetHeader from property_key
            property_key = property_key.replace("SetHeader", "", 1)

		# Check if the first letter of Component name is in Uppercase
		# If Yes, Split the propertyKey based on Uppercase letters
        if(property_key[0].isupper() is False):
            raise Exception(Messages.WRONG_PROPERTY_FORMAT.format(actual_property_name))
        else:
            # Get the Component name string to be removed from the key    
            parts = re.findall(r'([A-Z][^A-Z]*)', property_key)
            discard_letter = parts[0]
        
		# Check if property name contains the header name that starts with upper case
		# If Yes, Remove the previously found Component Name to get the Header Name
        if len(parts)<=1 :
            raise Exception(Messages.WRONG_PROPERTY_FORMAT.format(actual_property_name))

        else:
            # Replace the first letter with the empty string.
            response_header = property_key.replace(discard_letter, "")

        return response_header



# *********************************************************************
# This Original Work is copyright of 51 Degrees Mobile Experts Limited.
# Copyright 2019 51 Degrees Mobile Experts Limited, 5 Charlotte Close,
# Caversham, Reading, Berkshire, United Kingdom RG4 7BY.
#
# This Original Work is licensed under the European Union Public Licence (EUPL)
# v.1.2 and is subject to its terms as set out below.
#
# If a copy of the EUPL was not distributed with this file, You can obtain
# one at https://opensource.org/licenses/EUPL-1.2.
#
# The 'Compatible Licences' set out in the Appendix to the EUPL (as may be
# amended by the European Commission) shall be deemed incompatible for
# the purposes of the Work and the provisions of the compatibility
# clause in Article 5 of the EUPL shall not apply.
#
# If using the Work as, or as part of, a network application, by
# including the attribution notice(s) required under Article 5 of the EUPL
# in the end user terms of the application under an appropriate heading,
# such notice(s) shall fulfill the requirements of that article.
# ********************************************************************

from .evidence import Evidence
import traceback


class FlowData:

    """!
    FlowData is created by a specific Pipeline
    It collects evidence set by the user
    It passes evidence to FlowElements in the Pipeline
    These elements can return ElementData or populate an errors object

    """

    def __init__(self, pipeline):

        """!
        FlowData constructor.

        @type pipeline: Pipeline
        @param pipeline: parent pipeline

        """

        self.data = {}
        self.errors = {}
        self.pipeline = pipeline
        self.processed = False
        self.stopped = False
        self.evidence = Evidence(self)

    def process(self):

        """!
        Runs the process function on every attached flowElement allowing data to be changed based on evidence.
        This can only be run once per FlowData instance.

        @rtype: FlowData
        @return: Returns flowdata

        """

        if not self.processed:

            for flow_element in self.pipeline.flow_elements:
                if self.stopped is not True:
                    # All errors are caught and stored in an errors array keyed by the
                    # flowElement that set the error

                    try:
                        flow_element.process(self)

                    except Exception:

                        self.set_error(flow_element.datakey, traceback.format_exc())

            # Set processed flag to true. flowdata can only be processed once

            self.processed = True
            return self

        else:
            self.setError("error", "FlowData already processed")


    def get_from_element(self, flow_element):

        """!
        Retrieve data by FlowElement object.

        @type flow_element: FlowElement
        @param flow_element: FlowElement that created the data of interest

        @rtype: ElementData
        @return: Returns data that was created by the flowElement held in the FlowData

        """

        try:
            return self.get(flow_element.datakey)

        except Exception:
            return None


    def get(self, flow_element_key):

        """!
        Retrieve data by flowElement key.
        Called by FlowData.get_from_element method.

        @type flow_element_key: string
        @param flow_element_key: FlowElement.datakey of the FlowElement that created the data of interest

        @rtype: ElementData
        @return: Returns data in the FlowData instance that is under the specified key

        """

        try:
            return self.data[flow_element_key.lower()]

        except Exception:

            return self.pipeline.element_not_found(flow_element_key, self)

    def __getattr__(self, flow_element_key):

        """!
        Magic getter to allow retrieval of data from FlowData.data[flowElementKey] by flowElement name.
        For example, instead of `flowdata.get("device")` you can use `flowdata.device`

        @type flow_element_key: string
        @param flow_element_key: datakey of the FlowElement that created the data of interest

        @rtype: ElementData
        
        """

        return self.get(flow_element_key)

    def __getitem__(self, flow_element_key):

        """!
        Magic method in Python, which when used in a class, allows its instances to use the [] (indexer) operators.
        For example, instead of `flowdata.get("device")` you can use `flowdata["device"]`

        @type flow_element_key: string
        @param flow_element_key: datakey of the FlowElement that created the data of interest

        @rtype: ElementData
        
        """

        return self.get(flow_element_key)
		
    def set_element_data(self, element_data):

        """!
        Set data (used by flowElement) within FlowData.data

        @type element_data: ElementData
        @param element_data: elementData to be added to flowdata

        """

        self.data[element_data.flow_element.datakey] = element_data


    def set_error(self, key, error):

        """!
        Set error (should be keyed by flowElement datakey)

        @type key: string
        @param key: a flowElement.datakey

        @type error: string
        @param error: Error message

        """

        if key not in self.errors:
            self.errors[key] = list()

        self.errors[key].append(error)

        self.pipeline.log("error", error)


    def get_evidence_datakey(self):

        """!
        Get a list of evidence stored in the flowdata, filtered by
        its flowElements' evidenceKeyFilters

        @rtype: list
        @return: Returns filtered evidence

        """
        requestedEvidence = list()

        for flow_element in self.pipeline.flow_elements:
            requested_evidence = requestedEvidence.extend(flow_element.filter_evidence(self))

        return requested_evidence


    def stop(self):

        """!
        Stop processing any subsequent flowElements
        @return void
        
        """

        self.stopped = True

    def get_where(self, metakey, metavalue):

        """!
        Get data from flowElement based on property meta data

        @type metakey: str
        @param metakey: metakey shared by all data of interest

        @type metavalue: mixed
        @param metavalue: meta value shared by all data of interest

        @rtype: dict
        @return: Returns dictionary of data created by the flowElement that have the specified meta property

        """

        meta_query_output = {}

        properties = self.pipeline.get_properties()

        for flow_element, flow_element_properties in properties.items():
            for propertykey, propertymeta in flow_element_properties.items():
                if metakey.lower() in propertymeta:
                    if propertymeta[metakey.lower()] == metavalue:
                        try:
                            meta_query_output[propertykey] = self.get(flow_element).get(propertykey)                
                        # We are ignoring errors in getWhere as properties could be missing on purpose
                        # They shouldn't throw an error breaking the whole getWhere.
                        except Exception:
                            pass
                        
        return meta_query_output
                        
